//
//  TDSConnection.m
//  SQLKit
//
//  Created by Richard Penwell on 2/1/10.
//  Copyright 2010 Richard Penwell. All rights reserved.
//

#import "TDSConnection.h"
#import "NSOutputStream+Extensions.h"

const NSInteger kTDSDefaultPort = 1433;
const NSTimeInterval kTDSDefaultConnectionTimeout = 30.0;
const NSInteger kTDSConnectionTimeoutError = 1;
NSString* kTDSConnectionErrorDomain = @"TDSConnectionErrorDomain";

@interface TDSConnection ()

- (void)handleInputStreamEvent:(NSStreamEvent)eventCode;
- (void)handleOutputStreamEvent:(NSStreamEvent)eventCode;
- (void)openComplete;
- (void)handleIncomingData;
- (void)recievePacket:(TDSHeader)header withData:(NSData*)packetData;

@end

@implementation TDSConnection

@synthesize port, host, instanceName, databaseName, username, password, delegate;

+ (TDSConnection *)connectionWithHost:(NSHost *)host
{
	return [[[TDSConnection alloc] initWithHost:host] autorelease];
}

+ (TDSConnection *)connectionWithHost:(NSHost *)host andPort:(NSInteger)port
{
	return [[[TDSConnection alloc] initWithHost:host andPort:port] autorelease];
}

- (TDSConnection*)init
{
	[self doesNotRecognizeSelector:_cmd];
	[self release];
	return nil;
}

- (TDSConnection *)initWithHost:(NSHost *)_host
{
	return [self initWithHost:_host andPort:kTDSDefaultPort];
}

- (TDSConnection *)initWithHost:(NSHost *)_host andPort:(NSInteger)_port
{
	if (self = [super init])
	{
		if (_host == nil)
		{
			@throw [NSException exceptionWithName:NSInvalidArgumentException 
										   reason:@"host must not be nil" 
										 userInfo:nil]; 
		}	
		
		[self setHost:_host];
		[self setPort:_port];
		
		pendingPackets = [NSMutableArray new];
		state = TDSConnectionStateNotOpen;
	}
	
	return self;
}

- (void)open
{
	[self openWithTimeout:kTDSDefaultConnectionTimeout];
}

- (void)openWithTimeout:(NSTimeInterval)timeout;
{
	[self close];
	
	
	[NSStream getStreamsToHost:host 
						  port:port 
				   inputStream:&inputStream 
				  outputStream:&outputStream];
	[inputStream retain];
	[outputStream retain];
	[inputStream setDelegate:self];
	[outputStream setDelegate:self];
	
	[inputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] 
						   forMode:NSDefaultRunLoopMode];
	[outputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] 
							forMode:NSDefaultRunLoopMode];
	
	[self performSelector:@selector(abortConnect) withObject:nil afterDelay:timeout];
	
	[inputStream open];
	[outputStream open];
	
	state = TDSConnectionStateOpening;
}

- (void)abortConnect
{
	if ([self streamStatus] == NSStreamStatusOpen) return;
	
	[self close];
	NSError* timeoutError = [NSError errorWithDomain:kTDSConnectionErrorDomain 
												code:kTDSConnectionTimeoutError 
											userInfo:nil];
	if ([delegate respondsToSelector:@selector(connection:didFailToOpenReason:)])
		[delegate connection:self didFailToOpenReason:timeoutError];
}

- (void)close
{
	if (inputStream)
	{
		[inputStream close];
		inputStream = nil;
	}
	
	if (outputStream)
	{
		[outputStream close];
		outputStream = nil;
	}
}

- (NSStreamStatus)streamStatus
{
	return (NSStreamStatus)([inputStream streamStatus] & 
							[outputStream streamStatus]);
}


- (void)stream:(NSStream *)stream handleEvent:(NSStreamEvent)eventCode
{
	if (stream == inputStream) {
		[self handleInputStreamEvent:eventCode];
	} else if (stream == outputStream) {
		[self handleOutputStreamEvent:eventCode];
	} else {
		NSLog(@"Recieved a stream event for instance %@ which was not the input or output stream", stream);
	}
}

- (void)writePacketData:(NSData *)data withTokenType:(TDSPacketType)type status:(TDSStatus)status
{
	NSMutableData* packet = [NSMutableData dataWithCapacity:([data length] + 8)];
	[packet appendByte:type];
	[packet appendByte:status];
	[packet appendUShort:(8 + [data length]) byteOrder:CFByteOrderBigEndian];
	[packet appendUShort:0 byteOrder:CFByteOrderBigEndian];
	[packet appendByte:packetNumber++];
	[packet appendByte:0];
	[packet appendData:data];
	
	if ([outputStream hasSpaceAvailable])
	{
		[outputStream writeData:packet];
	}
	else {
		[pendingPackets addObject:packet];
	}
}

- (void)dealloc
{
	if (inputStream) [inputStream release];
	if (outputStream) [outputStream release];
	if (host) [host release];
}

- (void)handleInputStreamEvent:(NSStreamEvent)eventCode {
	switch (eventCode) {
		case NSStreamEventOpenCompleted:
			if ([self streamStatus] == NSStreamStatusOpen) [self openComplete];
			break;
		case NSStreamEventHasBytesAvailable:
			[self handleIncomingData];
			break;
		default:
			break;
	}
}

- (void)handleOutputStreamEvent:(NSStreamEvent)eventCode {
	switch (eventCode) {
		case NSStreamEventOpenCompleted:
			if ([self streamStatus] == NSStreamStatusOpen) [self openComplete];
			break;
		case NSStreamEventHasSpaceAvailable:
			if ([pendingPackets count] > 0)
			{
				NSData* singlePacket = [pendingPackets objectAtIndex:0];
				[pendingPackets removeObject:singlePacket];
				[outputStream writeData:singlePacket];
			}
			break;

		default:
			break;
	}
}

- (void)openComplete
{
	if (state == TDSConnectionStateOpening)
	{
		state = TDSConnectionStatePreloginSent;
		core = [TDSCore new];
		[core setConnection:self];
		
		[core preLoginWithParameters:nil];
	}
}

- (void)handleIncomingData
{
	NSData* packet;
	
	if ([inputStream hasBytesAvailable])
	{
		// If there is a previous packet, append the data to it before processing.
		if (partialPacket == nil)
			packet = [inputStream readData];
		else {
			NSMutableData* temporaryPacket = [NSMutableData dataWithData:partialPacket];
			[temporaryPacket appendData:[inputStream readData]];
			packet = temporaryPacket;
		}
	}
	else {
		packet = partialPacket;
	}
	
	if (packet == nil) return;

	
	// Check that at a minimum we have a complete header
	if (packet != nil && [packet length] >= sizeof(TDSHeader))
	{
		// Read the data about this packet from the stream
		TDSHeader header;
		NSDataReader* reader = [NSDataReader readerWithData:packet];
		header.type = [reader readByte];
		header.status = [reader readByte];
		header.length = [reader readUShortWithByteOrder:CFByteOrderBigEndian];
		header.sessionIdentifier = [reader readUShortWithByteOrder:CFByteOrderBigEndian];
		header.sequanceNumber = [reader readByte];
		header.window = [reader readByte];
		
		NSUInteger packetLength = header.length;
		
		// If we have a complete packet
		if ([packet length] >= packetLength)
		{
			NSData* packetData = [packet subdataWithRange:(NSRange){8, header.length-8}];
			[self recievePacket:header withData:packetData];
			
			// If we have MORE then a complete packet
			if ([packet length] > packetLength)
			{
				partialPacket = [packet subdataWithRange:NSMakeRange(packetLength,
																	 ([packet length] - packetLength))];
				
				// If there is more data after this packet, call ourself to process the additional
				//  data on the stream as it may contain more packets.
				if (partialPacket != nil) [self handleIncomingData];
			}
		}
		else {
			// We have some data, but not enough to create the full packet.
			//  The connection will hang out like this until we have recieved
			//  more data from the steram, and if at that time we have all of
			//  the data, then we will process, or wait more.
			partialPacket = packet;
		}
	}
	// We dont have a complete header, just wait for more data to arrive
	else {
		partialPacket = packet;
	}
}

- (void)recievePacket:(TDSHeader)header withData:(NSData*)packetData {
	if (state == TDSConnectionStatePreloginSent &&
		header.type == TDSPacketTypePreLogin)
	{
		// We are to assume that the response is a end of message when
		//  we are in the prelogin negotiation phase.
		NSAssert(header.status & 1, @"Prelogin response should be marked as end of message");
		
		NSDictionary* serverEnvironment = [core processPreloginResponse:packetData];
		
		if ([serverEnvironment valueForKey:kTDSEncryptionType])
		{
			NSNumber* encryptionRequired = [serverEnvironment valueForKey:kTDSEncryptionType];
			if ([encryptionRequired charValue] == TDSEncryptionRequired)
			{
				@throw [NSException exceptionWithName:@"TDSEncryptionRequiredError"
											   reason:@"The server requires encryption which is not supported"
											 userInfo:nil];
			}
		}
		if ([serverEnvironment valueForKey:kTDSInstanceNameCorrect])
		{
			NSNumber* instanceNameCorrect = [serverEnvironment valueForKey:kTDSInstanceNameCorrect];
			if ([instanceNameCorrect boolValue])
			{
				@throw [NSException exceptionWithName:@"TDSWrongInstanceNameError"
											   reason:@"The instance name provided is incorrect" 
											 userInfo:nil];
			}
		}
		tdsServerVersion = [[serverEnvironment valueForKey:kTDSServerVersion] intValue];
		tdsServerBuild = [[serverEnvironment valueForKey:kTDSServerBuild] intValue];
		
		state = TDSConnectionStateLoginSent;
		[core loginToDatabaseName:databaseName username:username password:password];
	}
	if (header.type == TDSPacketTypeResponse)
	{
		if (resultPacket != nil) {
			[resultPacket appendAdditionalData:packetData];
		}
		else {
			resultPacket = [TDSResultPacket resultPacketWithDelegate:self data:packetData];
		}
		if (header.status & TDSStatusEndOfMessage) { [resultPacket completePacket]; resultPacket = nil; }
	}
}

@end
